/*-
 * Copyright (c) 2010 Kai Wang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <errno.h>
#include <libelftc.h>

#include "_libelftc.h"


#define	AR_CHECK_BACKEND_IMPL(X,B,R) do {		\
	if ((B) >= ELFTC_AB_MAX) {			\
		ELFTC_SET_ERROR(ARGUMENT, 0);		\
		return (R);				\
	}						\
	if (_ar_backend[(B)].X == NULL) {		\
		ELFTC_SET_ERROR(UNIMPL, 0);		\
		return (R);				\
	}						\
	} while (0)

#define	AR_CHECK_NULL_ARG(A,R) do {			\
	if ((A) == NULL) {				\
		ELFTC_SET_ERROR(ARGUMENT, 0);		\
		return (R);				\
	}						\
	} while (0)

#define	AR_CHECK_AR_OBJ(X,A,R) do {			\
	AR_CHECK_NULL_ARG(A,R);				\
	AR_CHECK_BACKEND_IMPL(X,(A)->ao_be,R);		\
	} while (0)

static struct _Ar_Backend_CB _ar_backend[ELFTC_AB_MAX] = {

	[ELFTC_AB_LIBELF] = {
		.open_fd	 = _elftc_ar_libelf_open_fd,
		.open_file	 = _elftc_ar_libelf_open_file,
		.open_memory	 = _elftc_ar_libelf_open_memory,
		.set_flavor	 = NULL,
		.set_compression = NULL,
		.next		 = _elftc_ar_libelf_next,
		.rand		 = _elftc_ar_libelf_rand,
		.rewind		 = _elftc_ar_libelf_rewind,
		.get_arsym	 = _elftc_ar_libelf_get_arsym,
		.get_arhdr	 = _elftc_ar_libelf_get_arhdr,
		.get_data	 = _elftc_ar_libelf_get_data,
		.get_elf	 = _elftc_ar_libelf_get_elf,
		.extract	 = NULL,
		.new_arhdr	 = NULL,
		.new_data	 = NULL,
		.close		 = _elftc_ar_libelf_close,
	},

	[ELFTC_AB_LIBELFTC] = {
		.open_fd	 = NULL,
		.open_file	 = NULL,
		.open_memory	 = NULL,
		.set_flavor	 = NULL,
		.set_compression = NULL,
		.next		 = NULL,
		.rand		 = NULL,
		.rewind		 = NULL,
		.get_arsym	 = NULL,
		.get_arhdr	 = NULL,
		.get_data	 = NULL,
		.get_elf	 = NULL,
		.extract	 = NULL,
		.new_arhdr	 = NULL,
		.new_data	 = NULL,
		.close		 = NULL,
	},

	[ELFTC_AB_LIBARCHIVE] = {
		.open_fd	 = _elftc_ar_libarchive_open_fd,
		.open_file	 = _elftc_ar_libarchive_open_file,
		.open_memory	 = _elftc_ar_libarchive_open_memory,
		.set_flavor	 = NULL,
		.set_compression = NULL,
		.next		 = _elftc_ar_libarchive_next,
		.rand		 = _elftc_ar_libarchive_rand,
		.rewind		 = _elftc_ar_libarchive_rewind,
		.get_arsym	 = NULL,
		.get_arhdr	 = _elftc_ar_libarchive_get_arhdr,
		.get_data	 = _elftc_ar_libarchive_get_data,
		.get_elf	 = _elftc_ar_libarchive_get_elf,
		.extract	 = NULL,
		.new_arhdr	 = NULL,
		.new_data	 = NULL,
		.close		 = _elftc_ar_libarchive_close,
	},
};

Ar_Obj *
elftc_ar_open_fd(int fd, Ar_Cmd cmd, Ar_Backend be, int flags)
{

	AR_CHECK_BACKEND_IMPL(open_fd, be, NULL);

	return (_ar_backend[be].open_fd(fd, cmd, flags));
}

Ar_Obj *
elftc_ar_open_file(const char *name, Ar_Cmd cmd, Ar_Backend be, int flags)
{

	AR_CHECK_BACKEND_IMPL(open_fd, be, NULL);

	return (_ar_backend[be].open_file(name, cmd, flags));
}

Ar_Obj *
elftc_ar_open_memory(void *image, size_t sz, Ar_Backend be, int flags)
{

	AR_CHECK_BACKEND_IMPL(open_fd, be, NULL);

	return (_ar_backend[be].open_memory(image, sz, flags));
}

int
elftc_ar_set_flavor(Ar_Obj *ao, Ar_Flavor af)
{

	AR_CHECK_AR_OBJ(set_flavor, ao, -1);

	return (_ar_backend[ao->ao_be].set_flavor(ao, af));
}

int
elftc_ar_set_compression(Ar_Obj *ao, Ar_Compress ac)
{

	AR_CHECK_AR_OBJ(set_compression, ao, -1);

	return (_ar_backend[ao->ao_be].set_compression(ao, ac));
}

int
elftc_ar_next(Ar_Obj *ao)
{

	AR_CHECK_AR_OBJ(next, ao, -1);

	return (_ar_backend[ao->ao_be].next(ao));
}

int
elftc_ar_rand(Ar_Obj *ao, off_t offset)
{

	AR_CHECK_AR_OBJ(rand, ao, -1);

	return (_ar_backend[ao->ao_be].rand(ao, offset));
}

int
elftc_ar_rewind(Ar_Obj *ao)
{

	AR_CHECK_AR_OBJ(rewind, ao, -1);

	return (_ar_backend[ao->ao_be].rewind(ao));
}

Elf_Arsym *
elftc_ar_get_arsym(Ar_Obj *ao, size_t *sz)
{

	AR_CHECK_AR_OBJ(get_arsym, ao, NULL);
	AR_CHECK_NULL_ARG(sz, NULL);

	return (_ar_backend[ao->ao_be].get_arsym(ao, sz));
}

Elf_Arhdr *
elftc_ar_get_arhdr(Ar_Obj *ao)
{

	AR_CHECK_AR_OBJ(get_arhdr, ao, NULL);

	return (_ar_backend[ao->ao_be].get_arhdr(ao));
}

void *
elftc_ar_get_data(Ar_Obj *ao, size_t *sz)
{

	AR_CHECK_AR_OBJ(get_data, ao, NULL);
	AR_CHECK_NULL_ARG(sz, NULL);

	return (_ar_backend[ao->ao_be].get_data(ao, sz));
}

Elf *
elftc_ar_get_elf(Ar_Obj *ao)
{

	AR_CHECK_AR_OBJ(get_elf, ao, NULL);

	return (_ar_backend[ao->ao_be].get_elf(ao));
}

int
elftc_ar_extract(Ar_Obj *ao, int eflags)
{

	AR_CHECK_AR_OBJ(extract, ao, -1);

	return (_ar_backend[ao->ao_be].extract(ao, eflags));
}

int
elftc_ar_new_arhdr(Ar_Obj *ao, Elf_Arhdr *hdr)
{

	AR_CHECK_AR_OBJ(new_arhdr, ao, -1);
	AR_CHECK_NULL_ARG(hdr, -1);

	return (_ar_backend[ao->ao_be].new_arhdr(ao, hdr));
}

int
elftc_ar_new_data(Ar_Obj *ao, void *data, size_t sz)
{

	AR_CHECK_AR_OBJ(new_data, ao, -1);
	AR_CHECK_NULL_ARG(data, -1);

	return (_ar_backend[ao->ao_be].new_data(ao, data, sz));
}

int
elftc_ar_close(Ar_Obj *ao)
{

	AR_CHECK_AR_OBJ(new_data, ao, -1);

	return (_ar_backend[ao->ao_be].close(ao));
}
