/*-
 * Copyright (c) 2010 Kai Wang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ar.h>
#include <archive.h>
#include <archive_entry.h>
#include <assert.h>
#include <libelf.h>
#include <libelftc.h>
#include <string.h>

#include "_libelftc.h"

#define	_AR_HDR_SIZE	60

struct _archive_handle {
	struct archive *ah_a;
	int ah_loaded;
};

static int
_elftc_ar_libarchive_new(Ar_Obj *ao)
{
	struct archive *a;
	struct _archive_handle *ah;

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ || ao->ao_cmd == ELFTC_AC_WRITE);

	ah = ao->ao_handle;

	if (ao->ao_cmd == ELFTC_AC_READ) {
		if ((a = archive_read_new()) == NULL) {
			ELFTC_SET_ERROR_MSG(LIBARCHIVE,
			    "archive_read_new failed", NULL);
			free(ah);
			return (-1);
		}
		archive_read_support_compression_all(a);
		archive_read_support_format_all(a);
	} else {
		if ((a = archive_write_new()) == NULL) {
			ELFTC_SET_ERROR_MSG(LIBARCHIVE,
			    "archive_write_new failed", NULL);
			free(ah);
			return (-1);
		}
		if (ao->ao_flv == ELFTC_AF_GNU)
			archive_write_set_format_ar_svr4(a);
		else
			archive_write_set_format_ar_bsd(a);
		archive_write_set_compression_none(a);
	}

	ah->ah_a = a;

	return (0);
}

static int
_elftc_ar_libarchive_end(Ar_Obj *ao)
{
	struct archive *a;

	assert(ao->ao_cmd == ELFTC_AC_READ || ao->ao_cmd == ELFTC_AC_WRITE);
	assert(ao->ao_handle != NULL);
	a = ((struct _archive_handle *)(ao->ao_handle))->ah_a;
	assert(a != NULL);

	if (ao->ao_cmd == ELFTC_AC_READ) {
		if (archive_read_close(a)) {
			ELFTC_SET_ARCHIVE_ERROR("archive_read_close");
			return (-1);
		}
		archive_read_finish(a);
	} else {
		if (archive_write_close(a)) {
			ELFTC_SET_ARCHIVE_ERROR("archive_write_close");
			return (-1);
		}
		archive_write_finish(a);
	}

	return (0);
}

static Ar_Obj *
_elftc_ar_libarchive_open(Ar_Cmd cmd, int flags)
{
	struct _archive_handle *ah;
	Ar_Obj *ao;

	if ((ah = calloc(1, sizeof(*ah))) == NULL) {
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}

	if ((ao = _elftc_ar_alloc_object(cmd, ELFTC_AB_LIBARCHIVE, flags)) ==
	    NULL) {
		free(ah);
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}

	ao->ao_handle = ah;

	if (_elftc_ar_libarchive_new(ao) < 0) {
		free(ah);
		free(ao);
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}

	return (ao);
}

int
_elftc_ar_libarchive_close(Ar_Obj *ao)
{
	struct _Ar_Member *am, *tam;

	assert(ao != NULL && ao->ao_handle != NULL);

	TAILQ_FOREACH_SAFE(am, &ao->ao_amlist, am_next, tam) {
		TAILQ_REMOVE(&ao->ao_amlist, am, am_next);
		if (am->am_elf != NULL)
			(void) elf_end(am->am_elf);
		if (am->am_hdr != NULL) {
			if (am->am_hdr != NULL) {
				if (am->am_hdr->ar_name != NULL)
					free(am->am_hdr->ar_name);
				free(am->am_hdr);
			}
		}
		if (am->am_data != NULL)
			free(am->am_data);
		free(am);
	}

	(void) _elftc_ar_libarchive_end(ao);
	free(ao->ao_handle);
	_elftc_ar_dealloc_object(ao);

	return (0);
}

static int
_elftc_ar_libarchive_check_arsym(Ar_Obj *ao, struct archive *a,
    struct archive_entry *entry)
{
	const char *name;
	char *buff;
	size_t size;

	(void) ao;

	name = archive_entry_pathname(entry);

	if (!strcmp(name, "/") || !strcmp(name, "__.SYMDEF") ||
	    !strcmp(name, "#_LLVM_SYM_TAB_#")) {
		size = archive_entry_size(entry);
		if (size == 0)
			return (1);
		if ((buff = malloc(size)) == NULL) {
			ELFTC_SET_ERROR(RESOURCE, 0);
			return (-1);
		}
		if (archive_read_data(a, buff, size) != (ssize_t) size) {
			ELFTC_SET_ARCHIVE_ERROR("archive_read_data");
			free(buff);
			return (-1);
		}
#if 0
		if (!strcmp(name, "/"))
			_elftc_ar_parse_gnu_arsym(ao, buff, size);
		else if (!strcmp(name, "__.SYMDEF"))
			_elftc_ar_parse_bsd_arsym(ao, buff, size);
		else
			_elftc_ar_parse_llvm_arsym(ao, buff, size);
#endif
		free(buff);
		return (1);
	}

	return (0);
}

#define	DEF_BLKSZ 10240

Ar_Obj *
_elftc_ar_libarchive_open_fd(int fd, Ar_Cmd cmd, int flags)
{
	struct _archive_handle *ah;
	struct archive *a;
	Ar_Obj *ao;

	if ((ao = _elftc_ar_libarchive_open(cmd, flags)) == NULL)
		return (NULL);

	assert(ao->ao_handle != NULL);
	ah = ao->ao_handle;
	a = ah->ah_a;

	if (archive_read_open_fd(ah->ah_a, fd, DEF_BLKSZ)) {
		ELFTC_SET_ARCHIVE_ERROR("archive_read_open_fd");
		_elftc_ar_libarchive_close(ao);
		return (NULL);
	}

	ao->ao_off = SARMAG;

	return (ao);
}

Ar_Obj *
_elftc_ar_libarchive_open_file(const char *name, Ar_Cmd cmd, int flags)
{
	struct _archive_handle *ah;
	struct archive *a;
	Ar_Obj *ao;

	if ((ao = _elftc_ar_libarchive_open(cmd, flags)) == NULL)
		return (NULL);

	assert(ao->ao_handle != NULL);
	ah = ao->ao_handle;
	a = ah->ah_a;

	if (archive_read_open_filename(ah->ah_a, name, DEF_BLKSZ)) {
		ELFTC_SET_ARCHIVE_ERROR("archive_read_open_file");
		_elftc_ar_libarchive_close(ao);
		return (NULL);
	}

	ao->ao_off = SARMAG;

	return (ao);
}

Ar_Obj *
_elftc_ar_libarchive_open_memory(void *image, size_t sz, int flags)
{
	struct _archive_handle *ah;
	struct archive *a;
	Ar_Obj *ao;

	if ((ao = _elftc_ar_libarchive_open(ELFTC_AC_READ, flags)) == NULL)
		return (NULL);

	assert(ao->ao_handle != NULL);
	ah = ao->ao_handle;
	a = ah->ah_a;

	if (archive_read_open_memory(ah->ah_a, image, sz)) {
		ELFTC_SET_ARCHIVE_ERROR("archive_read_open_file");
		_elftc_ar_libarchive_close(ao);
		return (NULL);
	}

	ao->ao_off = SARMAG;

	return (ao);
}

static int
_elftc_ar_libarchive_load_all(Ar_Obj *ao)
{

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);
	assert((ao->ao_flags & ELFTC_AR_SEQUENTIAL) == 0);

	while (_elftc_ar_libarchive_next(ao) == 0)
		;

	/* TODO check error code here. */
	return (0);
}

int
_elftc_ar_libarchive_next(Ar_Obj *ao)
{
	struct _Ar_Member *am;
	struct _archive_handle *ah;
	struct archive *a;
	struct archive_entry *entry;
	const char *name;
	size_t size;
	int r;

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);
	ah = ao->ao_handle;
	a = ah->ah_a;

	/*
	 * Do not call libarchive functions if all members are already
	 * loaded.
	 */
	if ((ao->ao_flags & ELFTC_AR_SEQUENTIAL) == 0 && ah->ah_loaded) {
		if (ao->ao_pos != NULL)
			ao->ao_pos = TAILQ_NEXT(ao->ao_pos, am_next);
		if (ao->ao_pos == NULL)
			return (-1);
		return (0);
	}

	for (;;) {
		r = archive_read_next_header(a, &entry);
		if (r == ARCHIVE_RETRY)
			continue;
		if (r == ARCHIVE_EOF) {
			ah->ah_loaded = 1;
			return (-1);
		}
		if (r != ARCHIVE_OK) {
			ELFTC_SET_ARCHIVE_ERROR("archive_read_next_hreader");
			return (-1);
		}

		/* If in SEQUENTIAL mode, free the current data and header first. */
		if (ao->ao_flags & ELFTC_AR_SEQUENTIAL) {
			if (ao->ao_pos != NULL) {
				am = ao->ao_pos;
				if (am->am_hdr != NULL) {
					if (am->am_hdr->ar_name != NULL)
						free(am->am_hdr->ar_name);
					free(am->am_hdr);
				}
				if (am->am_data != NULL)
					free(am->am_data);
				ao->ao_pos = NULL;
			}
		}

		name = archive_entry_pathname(entry);
		size = archive_entry_size(entry);

		/* Skip GNU/SVR4 string table. */
		if (!strcmp(name, "//"))
			continue;

		/* Check if this member is symbol table. */
		r = _elftc_ar_libarchive_check_arsym(ao, a, entry);
		if (r < 0)
			return (-1);
		else if (r > 0)
			continue;

		/*
		 * Normal members otherwise. Allocate _Ar_Member storage
		 * for this member.
		 */
		if ((am = calloc(1, sizeof(*am))) == NULL) {
			ELFTC_SET_ERROR(RESOURCE, 0);
			return (-1);
		}
		if ((am->am_hdr = calloc(1, sizeof(*am->am_hdr))) == NULL) {
			ELFTC_SET_ERROR(RESOURCE, 0);
			return (-1);
		}

		/*
		 * Save the current member header.
		 */
		if ((am->am_hdr->ar_name = strdup(name)) == NULL) {
			ELFTC_SET_ERROR(RESOURCE, 0);
			free(am->am_hdr);
			am->am_hdr = NULL;
			return (-1);
		}
		am->am_hdr->ar_rawname = NULL;
		am->am_hdr->ar_date = archive_entry_mtime(entry);
		am->am_hdr->ar_uid = archive_entry_uid(entry);
		am->am_hdr->ar_gid = archive_entry_gid(entry);
		am->am_hdr->ar_mode = archive_entry_mode(entry);
		am->am_hdr->ar_size = size;

		/* Calculate the next member offset. */
		am->am_off = ao->ao_off;
		ao->ao_off += _AR_HDR_SIZE + size;
		if (ao->ao_off & 1)
			ao->ao_off++;

		ao->ao_pos = am;

		if (ao->ao_flags & ELFTC_AR_SEQUENTIAL)
			return (0);

		/*
		 * If ELFTC_AR_SEQUENTIAL is not set, read in the member
		 * content now.
		 */
		if (size > 0) {
			if ((am->am_data = malloc(size)) == NULL) {
				ELFTC_SET_ERROR(RESOURCE, 0);
				free(am->am_hdr);
				am->am_hdr = NULL;
				free(am);
				return (-1);
			}
			if (archive_read_data(a, am->am_data, size) !=
			    (ssize_t) size) {
				ELFTC_SET_ARCHIVE_ERROR("archive_read_data");
				free(am->am_hdr);
				am->am_hdr = NULL;
				free(am->am_data);
				free(am);
				return (-1);
			}
		} else
			am->am_data = NULL;

		TAILQ_INSERT_TAIL(&ao->ao_amlist, am, am_next);

		return (0);
	}
}

int
_elftc_ar_libarchive_rand(Ar_Obj *ao, off_t off)
{
	struct _archive_handle *ah;

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);
	ah = ao->ao_handle;

	/*
	 * _rand is not supported if ELFTC_AR_SEQUENTIAL is set.
	 */
	if (ao->ao_flags & ELFTC_AR_SEQUENTIAL) {
		ELFTC_SET_ERROR(IO, 0);
		return (-1);
	}

	/*
	 * Read in all the members if we haven't done that yet.
	 */
	if (!ah->ah_loaded) {
		if (_elftc_ar_libarchive_load_all(ao) < 0)
			return (-1);
	}

	/* Check if it's actually a rewind operation. */
	if (off == SARMAG) {
		ao->ao_pos = TAILQ_FIRST(&ao->ao_amlist);
		return (0);
	}

	if (ao->ao_pos == NULL)
		ao->ao_pos = TAILQ_FIRST(&ao->ao_amlist);

	if (ao->ao_pos->am_off == off)
		return (0);

	if (ao->ao_pos->am_off < off) {
		while (ao->ao_pos != NULL && ao->ao_pos->am_off < off)
			ao->ao_pos = TAILQ_NEXT(ao->ao_pos, am_next);
	} else {
		while (ao->ao_pos != NULL && ao->ao_pos->am_off > off)
			ao->ao_pos = TAILQ_PREV(ao->ao_pos, am_head, am_next);
	}

	if (ao->ao_pos == NULL || ao->ao_pos->am_off != off)
		return (-1);

	return (0);
}

int
_elftc_ar_libarchive_rewind(Ar_Obj *ao)
{

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);

	/*
	 * _rewind is not supported if ELFTC_AR_SEQUENTIAL is set.
	 */
	if (ao->ao_flags & ELFTC_AR_SEQUENTIAL) {
		ELFTC_SET_ERROR(IO, 0);
		return (-1);
	}

	return (_elftc_ar_libarchive_rand(ao, SARMAG));
}

Elf_Arhdr *
_elftc_ar_libarchive_get_arhdr(Ar_Obj *ao)
{

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);

	if (ao->ao_pos == NULL) {
		ELFTC_SET_ERROR(SEQUENCE, 0);
		return (NULL);
	}

	return (ao->ao_pos->am_hdr);
}

void *
_elftc_ar_libarchive_get_data(Ar_Obj *ao, size_t *sz)
{
	struct _Ar_Member *am;
	struct _archive_handle *ah;
	struct archive *a;

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);
	ah = ao->ao_handle;
	a = ah->ah_a;

	if (ao->ao_pos == NULL) {
		ELFTC_SET_ERROR(SEQUENCE, 0);
		return (NULL);
	}
	am = ao->ao_pos;

	if (am->am_hdr->ar_size == 0) {
		*sz = 0;
		return (NULL);
	}
	*sz = am->am_hdr->ar_size;

	if (am->am_data == NULL && ao->ao_flags & ELFTC_AR_SEQUENTIAL) {
		assert(am->am_data == NULL);
		if ((am->am_data = malloc(*sz)) == NULL) {
			ELFTC_SET_ERROR(RESOURCE, 0);
			return (NULL);
		}
		if (archive_read_data(a, am->am_data, *sz) != (ssize_t) *sz) {
			ELFTC_SET_ARCHIVE_ERROR("archive_read_data");
			free(am->am_data);
			am->am_data = NULL;
			return (NULL);
		}
	}

	assert(am->am_data != NULL);

	return (am->am_data);
}

#if 0
int
_elftc_ar_libarchive_extract(Ar_obj *ao, int eflags)
{

	/* Call archive_read_extract if ELFTC_AR_SEQUENTIAL is set and
	   member is not compressed. */

	 /* Otherwise extract the member ourselves */

	 /* Or, We should always call archive_read_data/get member data and
	    extract it oursevles */
}
#endif

Elf *
_elftc_ar_libarchive_get_elf(Ar_Obj *ao)
{
	struct _Ar_Member *am;

	assert(ao != NULL && ao->ao_handle != NULL);
	assert(ao->ao_cmd == ELFTC_AC_READ);

	/*
	 * _get_elf is not allowd to call if ELFTC_AR_SEQUENTIAL is set.
	 */
	if (ao->ao_flags & ELFTC_AR_SEQUENTIAL) {
		ELFTC_SET_ERROR(IO, 0);
		return (NULL);
	}

	if (ao->ao_pos == NULL) {
		ELFTC_SET_ERROR(SEQUENCE, 0);
		return (NULL);
	}
	am = ao->ao_pos;

	am->am_elf = elf_memory(am->am_data, am->am_hdr->ar_size);
	if (am->am_elf == NULL) {
		ELFTC_SET_ELF_ERROR("elf_memory");
		return (NULL);
	}

	return (am->am_elf);
}
