/*-
 * Copyright (c) 2010 Kai Wang
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ar.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libelf.h>
#include <libelftc.h>

#include "_libelftc.h"

struct _elf_descr {
	Elf *ed_elf;
	STAILQ_ENTRY(_elf_descr) ed_next;
};

struct _elf_handle {
	int eh_fd;		  	/* ar(1) archvie file descriptor. */
	Elf *eh_ar;			/* ar(1) archive ELF descriptor. */
	struct _elf_descr *eh_ed;	/* current member ELF descrptor. */
	STAILQ_HEAD(, _elf_descr) eh_edlist; /* list of open ELF descriptor. */
};

Ar_Obj *
_elftc_ar_libelf_open_fd(int fd, Ar_Cmd cmd, int flags)
{
	struct _elf_handle *eh;
	Ar_Obj *ao;

	if (cmd != ELFTC_AC_READ) {
		ELFTC_SET_ERROR(UNIMPL, 0);
		return (NULL);
	}

	if ((eh = calloc(1, sizeof(*eh))) == NULL) {
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}
	STAILQ_INIT(&eh->eh_edlist);
	eh->eh_fd = fd;

	if ((ao = _elftc_ar_alloc_object(cmd, ELFTC_AB_LIBELF, flags)) ==
	    NULL) {
		free(eh);
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}

	if ((eh->eh_ar = elf_begin(fd, ELF_C_READ, NULL)) == NULL) {
		free(eh);
		_elftc_ar_dealloc_object(ao);
		ELFTC_SET_ELF_ERROR("elf_begin");
		return (NULL);
	}

	if (elf_kind(eh->eh_ar) != ELF_K_AR) {
		(void) elf_end(eh->eh_ar);
		free(eh);
		_elftc_ar_dealloc_object(ao);
		ELFTC_SET_ERROR(ARGUMENT, 0);
		return (NULL);
	}


	ao->ao_handle = eh;

	return (ao);
}

Ar_Obj *
_elftc_ar_libelf_open_file(const char *name, Ar_Cmd cmd, int flags)
{
	int fd;

	if (cmd != ELFTC_AC_READ) {
		ELFTC_SET_ERROR(UNIMPL, 0);
		return (NULL);
	}

	if ((fd = open(name, O_RDONLY)) < 0) {
		ELFTC_SET_ERROR(IO, errno);
		return (NULL);
	}

	return (_elftc_ar_libelf_open_fd(fd, cmd, flags));
}

Ar_Obj *
_elftc_ar_libelf_open_memory(void *image, size_t sz, int flags)
{
	Ar_Obj *ao;
	struct _elf_handle *eh;

	if ((eh = calloc(1, sizeof(*eh))) == NULL) {
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}
	STAILQ_INIT(&eh->eh_edlist);
	eh->eh_fd = -1;

	if ((ao = _elftc_ar_alloc_object(ELFTC_AC_READ, ELFTC_AB_LIBELF,
	    flags)) == NULL) {
		free(eh);
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (NULL);
	}

	if ((eh->eh_ar = elf_memory(image, sz)) == NULL) {
		free(eh);
		_elftc_ar_dealloc_object(ao);
		ELFTC_SET_ELF_ERROR("elf_memory");
		return (NULL);
	}

	if (elf_kind(eh->eh_ar) != ELF_K_AR) {
		(void) elf_end(eh->eh_ar);
		free(eh);
		_elftc_ar_dealloc_object(ao);
		ELFTC_SET_ERROR(ARGUMENT, 0);
		return (NULL);
	}

	ao->ao_handle = eh;

	return (ao);
}

int
_elftc_ar_libelf_next(Ar_Obj *ao)
{
	struct _elf_handle *eh;
	struct _elf_descr *ed;
	Elf *ar;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;
	ar = eh->eh_ar;

	if (eh->eh_ed != NULL) {
		if (elf_next(eh->eh_ed->ed_elf) != ELF_C_READ) {
			ELFTC_SET_ELF_ERROR("elf_next");
			return (-1);
		}
	}

	if ((ed = calloc(1, sizeof(*ed))) == NULL) {
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (-1);
	}

	if ((ed->ed_elf = elf_begin(eh->eh_fd, ELF_C_READ, eh->eh_ar)) ==
	    NULL) {
		free(ed);
		ELFTC_SET_ELF_ERROR("elf_begin");
		return (-1);
	}
	eh->eh_ed = ed;
	STAILQ_INSERT_TAIL(&eh->eh_edlist, ed, ed_next);

	return (0);
}

int
_elftc_ar_libelf_rand(Ar_Obj *ao, off_t off)
{
	struct _elf_handle *eh;
	struct _elf_descr *ed;
	Elf *ar;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;
	ar = eh->eh_ar;

	if (elf_rand(ar, off) != off) {
		ELFTC_SET_ELF_ERROR("elf_rand");
		return (-1);
	}

	if ((ed = calloc(1, sizeof(*ed))) == NULL) {
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (-1);
	}

	if ((ed->ed_elf = elf_begin(eh->eh_fd, ELF_C_READ, ar)) == NULL) {
		free(ed);
		ELFTC_SET_ELF_ERROR("elf_begin");
		return (-1);
	}
	eh->eh_ed = ed;
	STAILQ_INSERT_TAIL(&eh->eh_edlist, ed, ed_next);

	return (0);
}

int
_elftc_ar_libelf_rewind(Ar_Obj *ao)
{
	struct _elf_handle *eh;
	struct _elf_descr *ed;
	Elf *ar;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;
	ar = eh->eh_ar;

	if (elf_rand(ar, SARMAG) != SARMAG) {
		ELFTC_SET_ELF_ERROR("elf_rand");
		return (-1);
	}

	if ((ed = calloc(1, sizeof(*ed))) == NULL) {
		ELFTC_SET_ERROR(RESOURCE, 0);
		return (-1);
	}

	if ((ed->ed_elf = elf_begin(eh->eh_fd, ELF_C_READ, ar)) == NULL) {
		free(ed);
		ELFTC_SET_ELF_ERROR("elf_begin");
		return (-1);
	}
	eh->eh_ed = ed;
	STAILQ_INSERT_TAIL(&eh->eh_edlist, ed, ed_next);

	return (0);
}

Elf_Arsym *
_elftc_ar_libelf_get_arsym(Ar_Obj *ao, size_t *sz)
{
	Elf *ar;
	Elf_Arsym *arsym;

	assert(ao != NULL && ao->ao_handle != NULL && sz != NULL);
	ar = ((struct _elf_handle *) ao->ao_handle)->eh_ar;

	if ((arsym = elf_getarsym(ar, sz)) == NULL) {
		ELFTC_SET_ELF_ERROR("elf_getarsym");
		return (NULL);
	}

	return (arsym);
}

Elf_Arhdr *
_elftc_ar_libelf_get_arhdr(Ar_Obj *ao)
{
	struct _elf_handle *eh;
	Elf_Arhdr *arh;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;
	
	if (eh->eh_ed == NULL) {
		ELFTC_SET_ERROR(SEQUENCE, 0);
		return (NULL);
	}

	if ((arh = elf_getarhdr(eh->eh_ed->ed_elf)) == NULL) {
		ELFTC_SET_ELF_ERROR("elf_getarhdr");
		return (NULL);
	}

	return (arh);
}

void *
_elftc_ar_libelf_get_data(Ar_Obj *ao, size_t *sz)
{
	struct _elf_handle *eh;
	char *data;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;

	if (eh->eh_ed == NULL) {
		ELFTC_SET_ERROR(SEQUENCE, 0);
		return (NULL);
	}

	if ((data = elf_rawfile(eh->eh_ed->ed_elf, sz)) == NULL) {
		ELFTC_SET_ELF_ERROR("elf_rawfile");
		return (NULL);
	}

	return (data);
}

Elf *
_elftc_ar_libelf_get_elf(Ar_Obj *ao)
{
	struct _elf_handle *eh;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;
	
	if (eh->eh_ed == NULL) {
		ELFTC_SET_ERROR(SEQUENCE, 0);
		return (NULL);
	}

	return (eh->eh_ed->ed_elf);
}

int
_elftc_ar_libelf_close(Ar_Obj *ao)
{
	struct _elf_handle *eh;
	struct _elf_descr *ed, *ted;

	assert(ao != NULL && ao->ao_handle != NULL);
	eh = ao->ao_handle;

	STAILQ_FOREACH_SAFE(ed, &eh->eh_edlist, ed_next, ted) {
		STAILQ_REMOVE(&eh->eh_edlist, ed, _elf_descr, ed_next);
		(void) elf_end(ed->ed_elf);
		free(ed);
	}
	(void) elf_end(eh->eh_ar);

	_elftc_ar_dealloc_object(ao);

	return (0);
}
